#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <inttypes.h>
#include <limits.h>

#include "linux_endian.h"
#include <linux/gfs_ondisk.h>
#define __user
#include <linux/gfs_ioctl.h>

/*  Constants  */
#define REDHAT_COPYRIGHT ("Copyright (C) Red Hat, Inc.  2004-2006  All rights reserved.")
#define OPTION_STRING ("f:t:Vh")
#ifndef TRUE
#define TRUE (1)
#endif

#ifndef FALSE
#define FALSE (0)
#endif

#define die(fmt, args...) \
do { \
	fprintf(stderr, "%s: ", prog_name); \
	fprintf(stderr, fmt, ##args); \
	exit(EXIT_FAILURE); \
} while (0)

#define GQ_OP_DUMP           (12)
#define GQ_OP_COPY           (13)
#define GQ_OP_STAT           (14)
#define GQ_OP_TRUNC          (15)

struct commandline {
	unsigned int operation;
	uint64_t trunc;
	int trunc_set;
	char file[PATH_MAX];
	char filesystem[PATH_MAX];
};
typedef struct commandline commandline_t;

char *prog_name;

/**
 * print_usage - print usage info to the user
 *
 */

static void
print_usage()
{
	printf("Usage:\n");
	printf("\n");
	printf("%s <dump|copy|stat|trunc> [options] <gfs mountpoint>\n",
	       prog_name);
	printf("\n");
	printf("Actions:\n");
	printf("  dump             dump gfs quota file to -f file\n");
	printf("  copy             copy -f file as gfs quota file\n");
	printf("  stat             print a stat of the gfs quota file\n");
	printf("  trunc            truncate gfs quota file to -t bytes\n");
	printf("\n");
	printf("Options:\n");
	printf("  -f               filename to dump/copy from\n");
	printf("  -t               bytes to truncate to\n");
	printf("  -V               Print program version information, then exit\n");
}

/**
 * check_for_gfs - Check to see if a descriptor is a file on a GFS filesystem
 * @fd: the file descriptor
 * @path: the path used to open the descriptor
 *
 */

void
check_for_gfs(int fd, char *path)
{
	unsigned int magic = 0;
	int error;

	error = ioctl(fd, GFS_IOCTL_IDENTIFY, &magic);
	if (error || magic != GFS_MAGIC)
		die("%s is not a GFS file/filesystem\n", path);
}

/**
 * decode_arguments - parse command line arguments
 * @argc: well, it's argc...
 * @argv: well, it's argv...
 * @comline: the structure filled in with the parsed arguments
 *
 * Function description
 *
 * Returns: what is returned
 */

static void
decode_arguments(int argc, char *argv[], commandline_t *comline)
{
	int cont = TRUE;
	int optchar;

	while (cont) {
		optchar = getopt(argc, argv, OPTION_STRING);

		switch (optchar) {
		case 'f':
			strcpy(comline->file, optarg);
			/*if (!realpath(optarg, comline->file))
				die("can't find %s: %s\n", optarg,
				strerror(errno));*/
			break;

		case 't':
			if (!isdigit(*optarg))
				die("argument to -t must be a number\n");
			sscanf(optarg, "%"SCNu64, &comline->trunc);
			comline->trunc_set = TRUE;
			break;

		case 'V':
			printf("gfs_quota_helper (built %s %s)\n", 
			       __DATE__, __TIME__);
			printf("%s\n", REDHAT_COPYRIGHT);
			exit(EXIT_SUCCESS);
			break;

		case 'h':
			print_usage();
			exit(EXIT_SUCCESS);
			break;

		case ':':
		case '?':
			fprintf(stderr, "Please use '-h' for usage.\n");
			exit(EXIT_FAILURE);
			break;

		case EOF:
			cont = FALSE;
			break;

		default:
			die("unknown option: %c\n", optchar);
			break;
		};
	}

	while (optind < argc) {
		if (strcmp(argv[optind], "dump") == 0) {
			if (comline->operation)
				die("can't specify two operations\n");
			comline->operation = GQ_OP_DUMP;
		} else if (strcmp(argv[optind], "copy") == 0) {
			if (comline->operation)
				die("can't specify two operations\n");
			comline->operation = GQ_OP_COPY;
		} else if (strcmp(argv[optind], "stat") == 0) {
			if (comline->operation)
				die("can't specify two operations\n");
			comline->operation = GQ_OP_STAT;
		} else if (strcmp(argv[optind], "trunc") == 0) {
			if (comline->operation)
				die("can't specify two operations\n");
			comline->operation = GQ_OP_TRUNC;
		} else if (!realpath(argv[optind], comline->filesystem)) {
			die("can't find %s: %s\n", optarg, strerror(errno));
		}
		optind++;
	}
	if (!(comline->operation && strlen(comline->filesystem) > 0))
		die("operation or filesystem unspecified. try -h for help\n");
}

void
do_dump(commandline_t *comline)
{
	int fd, fdout;
	struct gfs_ioctl gi;
	char buf[4096];
	uint64_t offset = 0;
	int error = 0;

	if (!*comline->file)
		die("need an output file to work on\n");

	if (!*comline->filesystem)
		die("need a filesystem to work on\n");
	
	fd = open(comline->filesystem, O_RDONLY);
	if (fd < 0)
		die("can't open file %s: %s\n", comline->filesystem,
		    strerror(errno));
	
	check_for_gfs(fd, comline->filesystem);

	fdout = open(comline->file, O_WRONLY | O_TRUNC | O_CREAT, 0600);
	if (fdout <0)
		die("can't open file %s: %s\n", comline->file, strerror(errno));

	do {
		char *argv[] = {"do_hfile_read", "quota"};
		
		gi.gi_argc = 2;
		gi.gi_argv = argv;
		gi.gi_data = buf;
		gi.gi_size = 4096;
		gi.gi_offset = offset;
		
		memset(buf, 0, 4096);
		
		error = ioctl(fd, GFS_IOCTL_SUPER, &gi);
		if (error < 0)
			die("can't read quota file: %s\n", strerror(errno));
		
		/* write to file */
		if (write(fdout, buf, error) != error)
			die("write failed to output file: %s\n", strerror(errno));
		
		offset += error;

	} while (error == 4096);
	
	close(fdout);
	close(fd);
}

void
do_copy(commandline_t *comline)
{
	int fd, fdin;
	struct gfs_ioctl gi;
	uint64_t offset = 0;
	char buf[4096];
	int error;
	
	if (!*comline->file)
		die("need an input file to work on\n");

	if (!*comline->filesystem)
		die("need a filesystem to work on\n");
	
	fd = open(comline->filesystem, O_RDONLY);
	if (fd < 0)
		die("can't open file %s: %s\n", comline->filesystem,
		    strerror(errno));
	
	check_for_gfs(fd, comline->filesystem);

	fdin = open(comline->file, O_RDONLY);
	if (fdin <0)
		die("can't open file %s: %s\n", comline->file, strerror(errno));
	
	{
		/* trunc the quota file to zero length */
		char *argv[] = {"do_hfile_trunc", "quota"};
		gi.gi_argc = 2;
		gi.gi_argv = argv;
		gi.gi_offset = 0;
		
		error = ioctl(fd, GFS_IOCTL_SUPER, &gi);
		if (error < 0)
			die("can't trunc quota file: %s\n", strerror(errno));
	}

	do {
		error = read(fdin, buf, 4096);
		if (error < 0)
			die("read failed on input file: %s\n", strerror(errno));
		
		{
			char *argv[] = {"do_hfile_write", "quota"};
			gi.gi_argc = 2;
			gi.gi_argv = argv;
			gi.gi_data = buf;
			gi.gi_size = error;
			gi.gi_offset = offset;

			error = ioctl(fd, GFS_IOCTL_SUPER, &gi);
			if (error < 0)
				die("can't stat quota file: %s\n", strerror(errno));
		}

		offset += error;
	} while (error == 4096);

	close(fdin);
	close(fd);
}

void
do_stat(commandline_t *comline)
{
	int fd;
	struct gfs_ioctl gi;
	char buf[sizeof(struct gfs_dinode)];
	struct gfs_dinode di;
	int error;
	
	if (!*comline->filesystem)
		die("need a filesystem to work on\n");
	
	fd = open(comline->filesystem, O_RDONLY);
	if (fd < 0)
		die("can't open file %s: %s\n", comline->filesystem,
		    strerror(errno));
	
	check_for_gfs(fd, comline->filesystem);

	{
		char *argv[] = {"get_hfile_stat", "quota"};
		gi.gi_argc = 2;
		gi.gi_argv = argv;
		gi.gi_data = buf;
		gi.gi_size = sizeof(struct gfs_dinode);

		memset(buf, 0, sizeof(struct gfs_dinode));
		
		error = ioctl(fd, GFS_IOCTL_SUPER, &gi);
		if (error < 0)
			die("can't stat quota file: %s\n", strerror(errno));

		memcpy(&di, buf, sizeof(struct gfs_dinode));
		fprintf(stdout, "----------------------------\n");
		fprintf(stdout, "  *** GFS QUOTA DINODE ***\n");
		fprintf(stdout, "----------------------------\n");
		gfs_dinode_print(&di);
	}

	close(fd);
}

void
do_trunc(commandline_t *comline)
{
	int fd;
	struct gfs_ioctl gi;
	int error;

	if (!comline->trunc_set)
		die("truncate value not set\n");

	if (!*comline->filesystem)
		die("need a filesystem to work on\n");
	
	fd = open(comline->filesystem, O_RDONLY);
	if (fd < 0)
		die("can't open file %s: %s\n", comline->filesystem,
		    strerror(errno));
	
	check_for_gfs(fd, comline->filesystem);

	{
		char *argv[] = {"do_hfile_trunc", "quota"};
		gi.gi_argc = 2;
		gi.gi_argv = argv;
		gi.gi_offset = comline->trunc;
		
		error = ioctl(fd, GFS_IOCTL_SUPER, &gi);
		if (error < 0)
			die("can't trunc quota file: %s\n", strerror(errno));
	}

	close(fd);
}

/**
 * main - Do everything
 * @argc: well, it's argc...
 * @argv: well, it's argv...
 *
 * Returns: exit status
 */

int
main(int argc, char *argv[])
{
	commandline_t comline;

	prog_name = argv[0];

	memset(&comline, 0, sizeof(commandline_t));

	decode_arguments(argc, argv, &comline);

	switch (comline.operation) {
	case GQ_OP_DUMP:
		do_dump(&comline);
		break;

	case GQ_OP_COPY:
		do_copy(&comline);
		break;

	case GQ_OP_STAT:
		do_stat(&comline);
		break;

	case GQ_OP_TRUNC:
		do_trunc(&comline);
		break;
	default:
		die("Unknown operation! try -h for help\n");
		break;
	}
	exit(EXIT_SUCCESS);
}
